==> kom%2B%2B-0.3.tar.gz: kom++-0.3/libconnection/connection.cc <==
// Handle a connection to a LysKOM server.
//
// $Id: connection.cc,v 1.23 1994/07/30 05:52:48 ceder Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <pwd.h>
#include <sys/param.h>
#include <unistd.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "connection.h"
#include "refno.h"
#include "question.h"
#include "async.h"
#include "get_info_q.h"

IscMaster *          Connection::mcb = NULL;
Connection *         Connection::the_one_and_only_connection = NULL;
SLList<IscEvent *>   Connection::external_events;

Connection::Connection(char *hostname, char *portnum, IscMaster *master)
    : server_info(), server_info_fetched(FALSE)
{
    IscConfig config;

    // +++Can not currently have more than one connection.
    assert(mcb == NULL);
    assert(the_one_and_only_connection == NULL);

    the_one_and_only_connection = this;
    hunt_nl_index = 0;
    str_skip_left = 0;
    strlen_gather = 0;
    sync_sink_data = NULL;
    async_sink = NULL;

    if (master != NULL)
	mcb = master;
    else
    {
	// Create a master.
	/*
	** Setup some parameters here
	*/
	config.version = 1005;
	config.master.version = 1001;
	config.master.memfn.alloc = NULL; // Use default allocators.
	config.master.memfn.realloc = NULL;
	config.master.memfn.free = NULL;
	config.master.abortfn = NULL; /* Use default abort function. */
#ifdef ISC_0_98_4
	config.session.version = 1002;
#else
	config.session.version = 1001;
#endif
	config.session.max.msgsize = -1; /* Use default sizes. */
	config.session.max.queuedsize = -1;
	config.session.max.dequeuelen = -1;
	config.session.max.openretries = -1;
	config.session.max.backlog = -1;
#ifdef ISC_0_98_4
	config.session.flags.txempty = 0; // No ISC_EVENT_EMPTY msgs, thanks!
#endif	
	mcb  = isc_initialize(&config);
	if ( mcb == NULL )
	{
	    cerr << "can't isc_initialize()\n";
	    return;
	}
    }

    server = isc_opentcp(mcb, hostname, portnum);
    if (server == NULL)
    {
	cerr << "can't connect to server.\n";
	return;
    }
    
    // Do protocol.

    char *name;
    name = getlogin();
    if (name == NULL)
    {
	struct passwd *ps = getpwuid(getuid());
	if (ps == NULL || ps->pw_name == NULL)
	{
	    static char buf[16];
	    sprintf(buf, "uid%ld", (long)getuid());
	    name = buf;
	}
	else
	    name = ps->pw_name;
    }

    // FIXME +++ configure
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif
    char hostbuf[MAXHOSTNAMELEN];
    char domainbuf[64];
    char dot[2];
    dot[0]='\0';
    dot[1]='\0';

    // FIXME: Solaris 2.3 says that this should not be used.
    if (gethostname(hostbuf, MAXHOSTNAMELEN) != 0)
	hostbuf[0] = '\0';

    if (strchr(hostbuf, '.') != NULL)
	domainbuf[0] = '\0';
    else
	if (getdomainname(domainbuf, 64) != 0) // FIXME: avoid this
	    domainbuf[0] = '\0';
	else
	    if (domainbuf[0] != '.')
		dot[0] = '.';

    int len = strlen(domainbuf);
    if (len > 0 && domainbuf[len-1] == '.')
	domainbuf[len] = '\0';

    LyStr me;
    me << name << '%' << hostbuf << dot << domainbuf;
    prot_a_LyStr helo("A");
    helo.append_hollerith(me);
    helo << "\r\n";
    send_question(helo);	// It isn't a question, but so what?
    LysKOM_gotten = 7;
}

Connection::~Connection()
{
    // Note: we do not delete these, since we do not own them.
    mcb = NULL;
    the_one_and_only_connection = NULL;
}

// Receive an event that is intended for this connection. This should
// never receive ISC_EVENT_TIMEOUT events, since they are handled in
// the callers of this method.
// Return values:
//   st_ok: A package was handled successfully
//   st_error: A unrecoverable error occured. Close this connection
//             and open a new one, or terminate the application.
//   st_pending: The event was a noop.
Status
Connection::receive_event(IscEvent *ev)
{
    switch(ev->event)
    {
    case ISC_EVENT_CONNECTED:
	return st_pending;
    case ISC_EVENT_REJECTED:
	return st_error;
    case ISC_EVENT_ERROR:
	return st_error;
    case ISC_EVENT_TIMEOUT:
    case ISC_EVENT_LOGIN:
    case ISC_EVENT_LOGOUT:
	abort();
    case ISC_EVENT_MESSAGE:
	unparsed.append(ev->msg->buffer, ev->msg->length);
	while (LysKOM_gotten && unparsed.strlen())
	{
	    assert(LysKOM_gotten == 1
		   || unparsed[0] == "MOKsyL"[LysKOM_gotten-2]);
	    LysKOM_gotten--;
	    unparsed.remove_first(1);
	}
	
	// Collect input in UNPARSED until we have an entire
	// packet (either a reply or an async message). The
	// end-of-packet is detected by looking for a newline that
	// is not part of a string.
	
	while (hunt_nl_index < unparsed.strlen())
	{
	    if (str_skip_left > 0)
	    {
		// We are currently in the middle of a string.
		
		if (str_skip_left + hunt_nl_index <= unparsed.strlen())
		{
		    hunt_nl_index += str_skip_left;
		    str_skip_left = 0;
		}
		else
		{
		    str_skip_left -= (unparsed.strlen() - hunt_nl_index);
		    hunt_nl_index = unparsed.strlen();
		}
	    }
	    else
	    {
		// We are not inside a string now.
		
		switch (unparsed[hunt_nl_index++])
		{
		case ' ':
		    // Skip whitespace.
		    strlen_gather = 0;
		    break;
		case 'H':
		    // We are entering a string.
		    str_skip_left = strlen_gather;
		    strlen_gather = 0;
		    break;
		case '0': case '1': case '2':
		case '3': case '4': case '5':
		case '6': case '7': case '8':
		case '9':
		    // We might be entering a string. Or, this may
		    // be any number (or even a bit string).
		    strlen_gather *= 10;
		    strlen_gather += unparsed[hunt_nl_index-1] - '0';
		    break;
		    
		case '{':
		   case '}':
		case ':':
		case '=':
		case '+':
		case '%':
		case '*':
		    // Some special characters that need no
		    // special attention.
		    break;
		    
		case '\n':
		    // We got an entire package. Handle it.
		    handle_package();
		    break;
		    
		default:
		    // Unexpected character.
		    cerr << "Protocol error!\n";
		    cerr << "Bad character: "
			 << unparsed[hunt_nl_index-1]
			 << " (" << (int)unparsed[hunt_nl_index-1]
			 << " decimal).\nPackage size: "
			 << unparsed.strlen()
			 << "\nIndex of error: "
			 << hunt_nl_index-1
			 << "\nThe packet: >"
			 << unparsed << "<\n";
		    assert(0);
		}
	    }
	}
	return st_ok;
    default:
	abort();
    }
}

Status
Connection::drain(int timeout)
{
    if (mcb == NULL || the_one_and_only_connection->server == NULL)
	return st_error;

    IscEvent *ev;
    int errs = 0;

    while ((ev = isc_getnextevent(mcb, timeout)) != NULL)
    {
	if (ev->event == ISC_EVENT_ERROR)
	{
	    // FIXME: Improve error handling.
	    fprintf(stderr, "isc_getnextevent error: %s\n", ev->msg->buffer);
	    sleep(1);
	    if (errs++ > 10)
		return st_error;
	}
	else
	{
	    if (ev->event == ISC_EVENT_TIMEOUT)
	    {
		isc_dispose(ev);
		return st_pending;
	    }
	    else if (ev->session == the_one_and_only_connection->server)
	    {
		// Need support for multiple servers here.
		Status retval = the_one_and_only_connection->receive_event(ev);
		isc_dispose(ev);
		return retval;
	    }
	    else
	    {
		assert(ev->session != NULL);
		timeout = 0;
		assert(external_events.OK());
		external_events.append(ev);
		assert(external_events.OK());
	    }
	}
    }

    return st_error;
}

IscEvent Connection::error_event;

IscEvent *
Connection::get_event(int timeout)
{
    if (mcb == NULL || the_one_and_only_connection->server == NULL)
	return &error_event;

    int errs = 0;

    assert(external_events.OK());
    if (!external_events.empty())
    {
	assert(external_events.OK());
	IscEvent *result = external_events.remove_front();
	assert(external_events.OK());
	return result;
    }

    IscEvent *ev;
    
    while ((ev = isc_getnextevent(mcb, timeout)) != NULL)
    {
	if (ev->event == ISC_EVENT_ERROR)
	{
	    fprintf(stderr, "isc_getnextevent error 2: %s\n", ev->msg->buffer);
	    sleep(1);
	    if (errs++ > 10)
		return &error_event;
	}
	else
	{
	    if (ev->event == ISC_EVENT_TIMEOUT)
	    {
		return ev;
	    }
	    else if (ev->session == the_one_and_only_connection->server)
	    {
		// Need support for multiple servers here.
		the_one_and_only_connection->receive_event(ev);
		isc_dispose(ev);
		return NULL;
	    }
	    else
	    {
		assert(ev->session != NULL);
		timeout = 0;
		return ev;
	    }
	}
    }
    
    return NULL;
}

void
Connection::handle_package()
{
    // Remember how much that should be left in unparsed when we have
    // handled this package.
    long rest = unparsed.strlen() - hunt_nl_index;
    
    if (unparsed[0] == ':')
    {
	unparsed.remove_first(1);
	hunt_nl_index--;
	if (async_sink != NULL)
	    (*async_sink)(sync_sink_data, unparsed.substr(hunt_nl_index));
    }
    else
    {
	int ok=0;
	switch(unparsed[0])
	{
	case '=':
	    ok=1;
	    break;
	case '%':
	    ok=0;
	    break;
	default:
	    abort();
	}
	
	unparsed.remove_first(1);
	int r = unparsed.stol();
	
	if (ok)
	{
	    assert(lookup_refno(r) != NULL);
	    lookup_refno(r)->parser(unparsed);
	}
	else
	{
	    int error = unparsed.stol();
	    // unparsed.stol();	// Skip the err_info.
	    assert(lookup_refno(r) != NULL);
	    lookup_refno(r)->receive_err(error);
	}
    }
    
    unparsed.remove_first(unparsed.strlen() - rest);
    hunt_nl_index = 0;
}

void
Connection::send_question(const prot_a_LyStr &str)
{
    if (mcb == NULL || server == NULL)
	return;

    char *p = str.malloc_copy();
    // FIXME check for 0
    isc_write(server, p, str.strlen());
    free(p);
}

void
Connection::register_async(void *vp, void (*new_sinc)(void *, prot_a_LyStr))
{
    sync_sink_data = vp;
    async_sink = new_sinc;
}


Server_info *
Connection::get_server_info()
{
    if (mcb == NULL || server == NULL)
	abort();		// +++ Error handling

    if (server_info_fetched == FALSE)
    {
	get_info_question   get_info_q(this);

	if (get_info_q.receive() == st_ok) 
	{
	    server_info = get_info_q.result();
	    server_info_fetched = TRUE;
	} 
	else 
	{
	    abort();		// +++Error handling
	}
    }
    
    return &server_info;
}

==> kom%2B%2B-0.4.tar.gz: kom++-0.4/libconnection/connection.cc <==
// Handle a connection to a LysKOM server.
//
// $Id: connection.cc,v 1.27 1995/11/18 22:48:53 ceder Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <pwd.h>
#include <sys/param.h>
#include <unistd.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <stdio.h>		// sometimes needed by <assert.h>.
#include "connection.h"
#include "refno.h"
#include "question.h"
#include "async.h"
#include "get_info_q.h"

IscMaster *          Connection::mcb = NULL;
Connection *         Connection::the_one_and_only_connection = NULL;
SLList<IscEvent *>   Connection::external_events;

Connection::Connection(char *hostname, char *portnum, IscMaster *master)
    : server_info(), server_info_fetched(false)
{
    IscConfig config;

    // +++Can not currently have more than one connection.
    assert(mcb == NULL);
    assert(the_one_and_only_connection == NULL);

    the_one_and_only_connection = this;
    hunt_nl_index = 0;
    str_skip_left = 0;
    strlen_gather = 0;
    sync_sink_data = NULL;
    async_sink = NULL;

    if (master != NULL)
	mcb = master;
    else
    {
	// Create a master.
	/*
	** Setup some parameters here
	*/
	config.version = 1005;
	config.master.version = 1001;
	config.master.memfn.alloc = NULL; // Use default allocators.
	config.master.memfn.realloc = NULL;
	config.master.memfn.free = NULL;
	config.master.abortfn = NULL; /* Use default abort function. */
#ifdef ISC_0_98_4
	config.session.version = 1002;
#else
	config.session.version = 1001;
#endif
	config.session.max.msgsize = -1; /* Use default sizes. */
	config.session.max.queuedsize = -1;
	config.session.max.dequeuelen = -1;
	config.session.max.openretries = -1;
	config.session.max.backlog = -1;
#ifdef ISC_0_98_4
	config.session.flags.txempty = 0; // No ISC_EVENT_EMPTY msgs, thanks!
#endif	
	mcb  = isc_initialize(&config);
	if ( mcb == NULL )
	{
	    cerr << "can't isc_initialize()\n";
	    return;
	}
    }

    server = isc_opentcp(mcb, hostname, portnum);
    if (server == NULL)
    {
	cerr << "can't connect to server.\n";
	return;
    }
    
    // Do protocol.

    char *name;
    name = getlogin();
    if (name == NULL)
    {
	struct passwd *ps = getpwuid(getuid());
	if (ps == NULL || ps->pw_name == NULL)
	{
	    static char buf[16];
	    sprintf(buf, "uid%ld", (long)getuid());
	    name = buf;
	}
	else
	    name = ps->pw_name;
    }

    // FIXME +++ configure
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif
    char hostbuf[MAXHOSTNAMELEN];
    char domainbuf[64];
    char dot[2];
    dot[0]='\0';
    dot[1]='\0';

    // FIXME: Solaris 2.3 says that this should not be used.
    if (gethostname(hostbuf, MAXHOSTNAMELEN) != 0)
	hostbuf[0] = '\0';

    if (strchr(hostbuf, '.') != NULL)
	domainbuf[0] = '\0';
    else
	if (getdomainname(domainbuf, 64) != 0) // FIXME: avoid this
	    domainbuf[0] = '\0';
	else
	    if (domainbuf[0] != '.')
		dot[0] = '.';

    int len = strlen(domainbuf);
    if (len > 0 && domainbuf[len-1] == '.')
	domainbuf[len] = '\0';

    LyStr me;
    me << name << '%' << hostbuf << dot << domainbuf;
    prot_a_LyStr helo("A");
    helo.append_hollerith(me);
    helo << "\r\n";
    send_question(helo);	// It isn't a question, but so what?
    LysKOM_gotten = 7;
}

Connection::~Connection()
{
    // Note: we do not delete these, since we do not own them.
    mcb = NULL;
    the_one_and_only_connection = NULL;
}

// Receive an event that is intended for this connection. This should
// never receive ISC_EVENT_TIMEOUT events, since they are handled in
// the callers of this method.
// Return values:
//   st_ok: A package was handled successfully
//   st_error: A unrecoverable error occured. Close this connection
//             and open a new one, or terminate the application.
//   st_pending: The event was a noop.
Status
Connection::receive_event(IscEvent *ev)
{
    switch(ev->event)
    {
    case ISC_EVENT_CONNECTED:
	return st_pending;
    case ISC_EVENT_REJECTED:
	return st_error;
    case ISC_EVENT_ERROR:
	return st_error;
    case ISC_EVENT_TIMEOUT:
    case ISC_EVENT_LOGIN:
    case ISC_EVENT_LOGOUT:
	abort();
    case ISC_EVENT_MESSAGE:
	unparsed.append(ev->msg->buffer, ev->msg->length);
	while (LysKOM_gotten && unparsed.strlen())
	{
	    assert(LysKOM_gotten == 1
		   || unparsed[0] == "MOKsyL"[LysKOM_gotten-2]);
	    LysKOM_gotten--;
	    unparsed.remove_first(1);
	}
	
	// Collect input in UNPARSED until we have an entire
	// packet (either a reply or an async message). The
	// end-of-packet is detected by looking for a newline that
	// is not part of a string.
	
	while (hunt_nl_index < unparsed.strlen())
	{
	    if (str_skip_left > 0)
	    {
		// We are currently in the middle of a string.
		
		if (str_skip_left + hunt_nl_index <= unparsed.strlen())
		{
		    hunt_nl_index += str_skip_left;
		    str_skip_left = 0;
		}
		else
		{
		    str_skip_left -= (unparsed.strlen() - hunt_nl_index);
		    hunt_nl_index = unparsed.strlen();
		}
	    }
	    else
	    {
		// We are not inside a string now.
		
		switch (unparsed[hunt_nl_index++])
		{
		case ' ':
		    // Skip whitespace.
		    strlen_gather = 0;
		    break;
		case 'H':
		    // We are entering a string.
		    str_skip_left = strlen_gather;
		    strlen_gather = 0;
		    break;
		case '0': case '1': case '2':
		case '3': case '4': case '5':
		case '6': case '7': case '8':
		case '9':
		    // We might be entering a string. Or, this may
		    // be any number (or even a bit string).
		    strlen_gather *= 10;
		    strlen_gather += unparsed[hunt_nl_index-1] - '0';
		    break;
		    
		case '{':
		case '}':
		case ':':
		case '=':
		case '+':
		case '%':
		case '*':
		    // Some special characters that need no
		    // special attention.
		    break;
		    
		case '\n':
		    // We got an entire package. Handle it.
		    handle_package();
		    break;
		    
		default:
		    // Unexpected character.
		    cerr << "Protocol error!\n";
		    cerr << "Bad character: "
			 << unparsed[hunt_nl_index-1]
			 << " (" << (int)unparsed[hunt_nl_index-1]
			 << " decimal).\nPackage size: "
			 << unparsed.strlen()
			 << "\nIndex of error: "
			 << hunt_nl_index-1
			 << "\nThe packet: >"
			 << unparsed << "<\n";
		    assert(0);
		}
	    }
	}
	return st_ok;
    default:
	abort();
    }
}

Status
Connection::drain(int timeout)
{
    if (mcb == NULL || the_one_and_only_connection->server == NULL)
	return st_error;

    IscEvent *ev;
    int errs = 0;

    while ((ev = isc_getnextevent(mcb, timeout)) != NULL)
    {
	if (ev->event == ISC_EVENT_ERROR)
	{
	    // FIXME: Improve error handling.
	    fprintf(stderr, "isc_getnextevent error: %s\n", ev->msg->buffer);
	    sleep(1);
	    if (errs++ > 10)
		return st_error;
	}
	else
	{
	    if (ev->event == ISC_EVENT_TIMEOUT)
	    {
		isc_dispose(ev);
		return st_pending;
	    }
	    else if (ev->session == the_one_and_only_connection->server)
	    {
		// Need support for multiple servers here.
		Status retval = the_one_and_only_connection->receive_event(ev);
		isc_dispose(ev);
		return retval;
	    }
	    else
	    {
		assert(ev->session != NULL);
		timeout = 0;
		assert(external_events.OK());
		external_events.append(ev);
		assert(external_events.OK());
	    }
	}
    }

    return st_error;
}

IscEvent Connection::error_event;

IscEvent *
Connection::get_event(int timeout)
{
    if (mcb == NULL || the_one_and_only_connection->server == NULL)
	return &error_event;

    int errs = 0;

    assert(external_events.OK());
    if (!external_events.empty())
    {
	assert(external_events.OK());
	IscEvent *result = external_events.remove_front();
	assert(external_events.OK());
	return result;
    }

    IscEvent *ev;
    
    while ((ev = isc_getnextevent(mcb, timeout)) != NULL)
    {
	if (ev->event == ISC_EVENT_ERROR)
	{
	    fprintf(stderr, "isc_getnextevent error 2: %s\n", ev->msg->buffer);
	    sleep(1);
	    if (errs++ > 10)
		return &error_event;
	}
	else
	{
	    if (ev->event == ISC_EVENT_TIMEOUT)
	    {
		return ev;
	    }
	    else if (ev->session == the_one_and_only_connection->server)
	    {
		// Need support for multiple servers here.
		the_one_and_only_connection->receive_event(ev);
		isc_dispose(ev);
		return NULL;
	    }
	    else
	    {
		assert(ev->session != NULL);
		return ev;
	    }
	}
    }
    
    return NULL;
}

void
Connection::handle_package()
{
    // Remember how much that should be left in unparsed when we have
    // handled this package.
    long rest = unparsed.strlen() - hunt_nl_index;
    
    if (unparsed[0] == ':')
    {
	unparsed.remove_first(1);
	hunt_nl_index--;
	if (async_sink != NULL)
	    (*async_sink)(sync_sink_data, unparsed.substr(hunt_nl_index));
    }
    else
    {
	int ok=0;
	switch(unparsed[0])
	{
	case '=':
	    ok=1;
	    break;
	case '%':
	    ok=0;
	    break;
	default:
	    abort();
	}
	
	unparsed.remove_first(1);
	int r = unparsed.stol();
	
	if (ok)
	{
	    assert(lookup_refno(r) != NULL);
	    lookup_refno(r)->parser(unparsed);
	}
	else
	{
	    int error = unparsed.stol();
	    // unparsed.stol();	// Skip the err_info.
	    assert(lookup_refno(r) != NULL);
	    lookup_refno(r)->receive_err(error);
	}
    }
    
    unparsed.remove_first(unparsed.strlen() - rest);
    hunt_nl_index = 0;
}

void
Connection::send_question(const prot_a_LyStr &str)
{
    if (mcb == NULL || server == NULL)
	return;

    char *p = str.malloc_copy();
    // FIXME check for 0
    isc_write(server, p, str.strlen());
    free(p);
}

void
Connection::register_async(void *vp, void (*new_sinc)(void *, prot_a_LyStr))
{
    sync_sink_data = vp;
    async_sink = new_sinc;
}


Server_info *
Connection::get_server_info()
{
    if (mcb == NULL || server == NULL)
	abort();		// +++ Error handling

    if (server_info_fetched == false)
    {
	get_info_question   get_info_q(this);

	if (get_info_q.receive() == st_ok) 
	{
	    server_info = get_info_q.result();
	    server_info_fetched = true;
	} 
	else 
	{
	    abort();		// +++Error handling
	}
    }
    
    return &server_info;
}
